"""Define constants with dataclasses for use in chasten."""

from dataclasses import dataclass
from pathlib import Path


# chasten constant
@dataclass(frozen=True)
class Chasten:
    """Define the Chasten dataclass for constant(s)."""

    Analyze_Storage: Path
    Application_Name: str
    Application_Author: str
    App_Storage: Path
    API_Key_Storage: Path
    Chasten_Database_View: str
    Emoji: str
    Executable_Fly: str
    Executable_Vercel: str
    Https: str
    Name: str
    Programming_Language: str
    Separator: str
    Server_Shutdown: str
    Tagline: str
    Theme_Background: str
    Theme_Colors: str
    Website: str


chasten = Chasten(
    Analyze_Storage=Path("analysis.md"),
    Application_Name="chasten",
    Application_Author="ChastenedTeam",
    App_Storage=Path("check.txt"),
    API_Key_Storage=Path("userapikey.txt"),
    Chasten_Database_View="chasten_complete",
    Emoji=":dizzy:",
    Executable_Fly="fly",
    Executable_Vercel="vercel",
    Https="https://",
    Name="chasten",
    Programming_Language="python",
    Separator="/",
    Server_Shutdown=":person_shrugging: Shut down chasten's sylog server",
    Tagline="chasten: Analyze the AST of Python Source Code",
    Theme_Background="default",
    Theme_Colors="ansi_dark",
    Website=":link: GitHub: https://github.com/gkapfham/chasten",
)


# checks constant
@dataclass(frozen=True)
class Checks:
    """Define the Checks dataclass for constant(s)."""

    Check_Chasten: str
    Check_Code: str
    Check_Count: str
    Check_Confidence: int
    Check_File: str
    Check_Id: str
    Checks_Label: str
    Check_Max: str
    Check_Min: str
    Check_Name: str
    Check_Pattern: str


checks = Checks(
    Check_Chasten="chasten",
    Check_Code="code",
    Check_Count="count",
    Check_Confidence=80,
    Check_File="checks-file",
    Check_Id="id",
    Checks_Label="checks",
    Check_Max="max",
    Check_Min="min",
    Check_Name="name",
    Check_Pattern="pattern",
)


# datasette constant
@dataclass(frozen=True)
class Datasette:
    """Define the Datasette dataclass for constant(s)."""

    Chasten_Database: str
    Datasette_Executable: str
    Datasette_Copyable_Install: str
    Datasette_Export_Notebook: str
    Datasette_Search_All: str


datasette = Datasette(
    Chasten_Database="chasten.db",
    Datasette_Executable="datasette",
    Datasette_Copyable_Install="--install=datasette-copyable",
    Datasette_Export_Notebook="--install=datasette-export-notebook",
    Datasette_Search_All="--install=datasette-search-all",
)


# filesystem constant
@dataclass(frozen=True)
class Filesystem:
    """Define the Filesystem dataclass for constant(s)."""

    Current_Directory: str
    Dash: str
    Dot: str
    Main_Configuration_File: str
    Main_Checks_File: str
    Main_Results_File_Name: str
    Main_Results_Combined_File_Name: str
    Main_Results_Flattened_Directory_Name: str
    Results_Extension: str


filesystem = Filesystem(
    Current_Directory=".",
    Dash="-",
    Dot=".",
    Main_Configuration_File="config.yml",
    Main_Checks_File="checks.yml",
    Main_Results_File_Name="chasten-results",
    Main_Results_Combined_File_Name="chasten-integrated-results",
    Main_Results_Flattened_Directory_Name="chasten-flattened-csvs-sqlite-db",
    Results_Extension="json",
)


# humanreadable constant
@dataclass(frozen=True)
class Humanreadable:
    """Define the Humanreadable dataclass for constant(s)."""

    Yes: str
    No: str


humanreadable = Humanreadable(Yes="Yes", No="No")


# logger constant
@dataclass(frozen=True)
class Logger:
    """Define the Logger dataclass for constant(s)."""

    Function_Prefix: str
    Richlog: str
    Syslog: str


logger = Logger(
    Function_Prefix="configure_logging_",
    Richlog="chasten-richlog",
    Syslog="chasten-syslog",
)


# logging constant
@dataclass(frozen=True)
class Logging:
    """Define the Logging dataclass for constant(s)."""

    Debug: str
    Info: str
    Warning: str
    Error: str
    Critical: str
    Console_Logging_Destination: str
    Syslog_Logging_Destination: str
    Default_Logging_Destination: str
    Default_Logging_Level: str
    Format: str
    Rich: str


logging = Logging(
    Debug="DEBUG",
    Info="INFO",
    Warning="WARNING",
    Error="ERROR",
    Critical="CRITICAL",
    Console_Logging_Destination="CONSOLE",
    Syslog_Logging_Destination="syslog",
    Default_Logging_Destination="console",
    Default_Logging_Level="ERROR",
    Format="%(message)s",
    Rich="Rich",
)


# markers constant
@dataclass(frozen=True)
class Markers:
    """Define the Markers dataclass for constant(s)."""

    Bad_Fifteen: str
    Bad_Zero_Zero: str
    Code_Context: int
    Comma_Space: str
    Empty_Bytes: bytes
    Empty_String: str
    Ellipse: str
    Forward_Slash: str
    Dot: str
    Hidden: str
    Indent: str
    Newline: str
    Non_Zero_Exit: int
    Nothing: str
    Single_Quote: str
    Slice_One: int
    Small_Bullet_Unicode: str
    Space: str
    Tab: str
    Underscore: str
    Xml: str
    Zero: int
    Zero_Exit: int
    Percent_Multiplier: int


markers = Markers(
    Bad_Fifteen="<15>",
    Bad_Zero_Zero="",
    Code_Context=5,
    Comma_Space=", ",
    Empty_Bytes=b"",
    Empty_String="",
    Ellipse="...",
    Forward_Slash="/",
    Dot=".",
    Hidden=".",
    Indent="   ",
    Newline="\n",
    Non_Zero_Exit=1,
    Nothing="",
    Single_Quote="'",
    Slice_One=1,
    Small_Bullet_Unicode="\u2022",
    Space=" ",
    Tab="\t",
    Underscore="_",
    Xml="xml",
    Zero=0,
    Zero_Exit=0,
    Percent_Multiplier=100,
)


# output constant
@dataclass(frozen=True)
class Output:
    """Define the Output dataclass for constant(s)."""

    Syslog: str
    Test_Start: str


output = Output(
    Syslog=":sparkles: Syslog server for receiving debugging information",
    Test_Start=":sparkles: Start to run test suite for the specified program",
)


# server constant
@dataclass(frozen=True)
class Server:
    """Define the Server dataclass for constant(s)."""

    Backup_Count: int
    Localhost: str
    Log_File: str
    Max_Log_Size: int
    Poll_Interval: float
    Port: int
    Utf8_Encoding: str


server = Server(
    Backup_Count=1,
    Localhost="127.0.0.1",
    Log_File=".discover.log",
    Max_Log_Size=1048576,
    Poll_Interval=0.5,
    Port=2525,
    Utf8_Encoding="utf-8",
)



"""Validate various aspects of configurations and command-line arguments."""

from typing import Any, Dict, List, Tuple

import jsonschema
from jsonschema.exceptions import ValidationError

from chasten import constants, output, util

# intuitive description:
# a configuration file links to one or more checks files
# see ./chasten in the root of the repository for the config.yml file
JSON_SCHEMA_CONFIG = {
    "type": "object",
    "required": [],
    "properties": {
        "chasten": {
            "type": "object",
            "properties": {
                "checks-file": {
                    "type": "array",
                    "items": {"type": "string"},
                    "required": [],
                },
            },
            "additionalProperties": False,
        },
    },
}

# intutive description:
# a checks file describes all of the details for one or more checks
# see ./chasten in the root of the repository for the checks.yml file
JSON_SCHEMA_CHECKS = {
    "type": "object",
    "properties": {
        "checks": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "id": {"type": "string"},
                    "description": {"type": "string"},
                    "pattern": {"type": "string"},
                    "code": {"type": "string"},
                    "count": {
                        "anyOf": [
                            {
                                "type": "object",
                                "properties": {
                                    "min": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ]
                                    }
                                },
                                "required": ["min"],
                            },
                            {
                                "type": "object",
                                "properties": {
                                    "max": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ]
                                    }
                                },
                                "required": ["max"],
                            },
                            {
                                "type": "object",
                                "properties": {
                                    "min": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ]
                                    },
                                    "max": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ]
                                    },
                                },
                                "required": ["min", "max"],
                            },
                        ]
                    },
                },
                "required": ["name", "id", "pattern", "code"],
                "additionalProperties": False,
            },
        }
    },
}


def extract_checks_file_name(
    configuration: Dict[str, Dict[str, Any]]
) -> Tuple[bool, List[str]]:
    """Validate the checks configuration."""
    # there is a main "chasten" key
    if constants.checks.Check_Chasten in configuration.keys():
        # there is a "checks-file" key
        if constants.checks.Check_File in configuration[constants.checks.Check_Chasten]:
            # extract the name of the checks-files
            # and return them in a list with a boolean
            # indicate to show that checks files were found
            checks_file_name_list = configuration[constants.checks.Check_Chasten][
                constants.checks.Check_File
            ]
            return (True, checks_file_name_list)
    # contents were not found and thus returen no filenames
    return (False, [constants.markers.Empty_String])


def validate_configuration(
    configuration: Dict[str, Dict[str, Any]],
    schema: Dict[str, Any] = JSON_SCHEMA_CONFIG,
) -> Tuple[bool, str]:
    """Validate the main configuration."""
    # indicate that validation passed; since there
    # were no validation errors, return an empty string
    try:
        jsonschema.validate(configuration, schema)
        return (True, constants.markers.Empty_String)
    # indicate that validation failed;
    # since validation errors exist, package them up
    # and return them along with the indication
    except ValidationError as validation_error:
        error_message = str(validation_error)
        error_message = error_message.lstrip()
        return (False, error_message)


def validate_checks_configuration(
    configuration: Dict[str, Dict[str, Any]]
) -> Tuple[bool, str]:
    """Validate the checks configuration."""
    return validate_configuration(configuration, JSON_SCHEMA_CHECKS)


def validate_file(
    file_name: str,
    configuration_file_yaml_str: str,
    yaml_data_dict: Dict[str, Dict[str, Any]],
    json_schema: Dict[str, Any] = JSON_SCHEMA_CONFIG,
    verbose: bool = False,
) -> bool:
    """Validate the provided file according to the provided JSON schema."""
    # perform the validation of the configuration file
    (validated, errors) = validate_configuration(yaml_data_dict, json_schema)
    output.console.print(
        f":sparkles: Validated {file_name}? {util.get_human_readable_boolean(validated)}"
    )
    # there was a validation error, so display the error report
    if not validated:
        output.console.print(f":person_shrugging: Validation errors:\n\n{errors}")
    # validation worked correctly, so display the configuration file
    else:
        output.opt_print_log(verbose, newline="")
        output.opt_print_log(verbose, label=f":sparkles: Contents of {file_name}:\n")
        output.opt_print_log(verbose, config_file=configuration_file_yaml_str)
    return validated



"""Extract and analyze details about specific checks."""

from typing import Dict, List, Tuple, Union

from chasten import constants, enumerations, util


def extract_min_max(
    check: Dict[str, Union[str, Dict[str, int]]]
) -> Tuple[Union[int, None], Union[int, None]]:
    """Extract the minimum and maximum values from the checks dictionary."""
    # extract information about the count attribute
    # and the min and max values if they exist
    min_count = check.get(constants.checks.Check_Count, {}).get(constants.checks.Check_Min)  # type: ignore
    max_count = check.get("count", {}).get("max")  # type: ignore
    return (min_count, max_count)


def extract_description(check: Dict[str, Union[str, Dict[str, int]]]) -> str:
    """Extract the description that may optionally be stored in a check."""
    # the attribute is not None and thus the function
    # should create the labelled attribute out of it
    if "description" in check:
        return str(check["description"])
    return ""


def create_attribute_label(attribute: Union[str, int, None], label: str) -> str:
    """Create an attribute label string for display as long as it is not null."""
    # define an empty attribute string, which is
    # the default in the case when attribute is None
    labeled_attribute = ""
    # the attribute is not None and thus the function
    # should create the labelled attribute out of it
    if attribute:
        labeled_attribute = f"{label} = {attribute}"
    return labeled_attribute


def join_attribute_labels(attribute_labels: List[str]) -> str:
    """Join all of the attribute labels in a comma-separated list."""
    # start the joined attribute labels with the empty string,
    # which is what it will be by default as well
    joined_attribute_labels = constants.markers.Empty_String
    # incrementally create the list of labelled attributes,
    # separating each one with a comma and a space and
    # ensuring that the last one does not have a trailing
    # comma and space after it
    for i, attribute_label in enumerate(attribute_labels):
        # only add the comma and the space when the for loop
        # is not dealing with the final value in the list of labels
        if i > 0:
            joined_attribute_labels += constants.markers.Comma_Space
        # append the new attribute label to the running list
        joined_attribute_labels += attribute_label  # type: ignore
    return joined_attribute_labels


def is_checkable(min_value: Union[int, None], max_value: Union[int, None]) -> bool:
    """Help to see if the value is in the closed interval."""
    if min_value is None and max_value is None:
        return False
    return True


def is_in_closed_interval(value: int, min_value: int, max_value: int) -> bool:
    """Help to see if the value is in the closed interval."""
    return min(max_value, value) == value and max(min_value, value) == value


def check_match_count(
    count: int, min_value: Union[int, None] = None, max_value: Union[int, None] = None
) -> bool:
    """Confirm that the count is between min_value and max_value."""
    # Overall description: if min_value is not None then count must be >= min_value.
    # If max_value is not None then count must be <= max_value
    # both of the values are None and thus the comparision is vacuously true
    if min_value is None and max_value is None:
        return True
    # both are not None and thus the count must be in the closed interval
    if min_value is not None and max_value is not None:
        return is_in_closed_interval(count, min_value, max_value)
    # at this point, only one of the values might not be None
    # if min_value is not None, then confirm that it is greater than or equal
    if min_value is not None:
        if count >= min_value:
            return True
    # if max_value is not None, then confirm that it is less than or equal
    if max_value is not None:
        if count <= max_value:
            return True
    # if none of those conditions were true, then the count is not
    # between the minimum and the maximum value, inclusively
    return False


def make_checks_status_message(check_status: bool) -> str:
    """Make a check status message in human readable format."""
    if check_status:
        return f":smiley: Did the check pass? {util.get_human_readable_boolean(check_status)}"
    return (
        f":worried: Did the check pass? {util.get_human_readable_boolean(check_status)}"
    )


def fix_check_criterion(
    criterion: Union[enumerations.FilterableAttribute, str, int]
) -> Union[str, int]:
    """Remove null values from a criterion."""
    # the converted criterion's default is an empty string
    new_criterion: Union[str, int] = ""
    # if the criterion is not None, then it should either be
    # the value in the enumeration or the value of a string or int
    if criterion is not None:
        # the criterion is an enum and thus the value must be extracted
        if type(criterion) is enumerations.FilterableAttribute:
            new_criterion = criterion.value
        # the criterion is not an enum and thus it must be
        # an int or a string that can be stored directly
        else:
            new_criterion = criterion  # type: ignore
    return new_criterion



"""Analyze the abstract syntax tree, its XML-based representation, and/or the search results."""

import json
from typing import Any, Dict, List, Tuple, Union

from pyastgrep import search as pyastgrepsearch  # type: ignore
from thefuzz import fuzz  # type: ignore

from chasten import constants, enumerations


def include_or_exclude_checks(
    checks: List[Dict[str, Union[str, Dict[str, int]]]],
    check_attribute: enumerations.FilterableAttribute,
    check_match: str,
    check_confidence: int = constants.checks.Check_Confidence,
    include: bool = True,
) -> List[Dict[str, Union[str, Dict[str, int]]]]:
    """Perform all of the includes and excludes for the list of checks."""
    filtered_checks = []
    # at least one aspect of the inputs was not specified (likely due to
    # the fact that the command-line argument(s) were not used) and thus
    # there is no filtering that should take place; return the input
    if check_attribute is None or check_match is None:
        return checks
    # the function's inputs are valid and so perform the filtering
    for check in checks:
        # extract the contents of the requested attribute for inclusion
        check_requested_include_attribute = check[check_attribute]
        # compute the fuzzy match value for the specific:
        # --> requested include attribute
        # --> specified match string
        fuzzy_value = fuzz.ratio(check_match, check_requested_include_attribute)
        # include the check if the fuzzy inclusion value is above (or equal to) threshold
        # and the purpose of the function call is to include values
        if (fuzzy_value >= check_confidence) and include:
            filtered_checks.append(check)
        # include the check if the fuzzy inclusion value is below threshold
        # and the purpose of the function call is to exclude values;
        # note that not including a value means that it excludes
        elif (fuzzy_value < check_confidence) and not include:
            filtered_checks.append(check)
    return filtered_checks


def filter_matches(
    match_list: List[Union[pyastgrepsearch.Match, Any]],
    data_type,
) -> Tuple[List[pyastgrepsearch.Match], List[Any]]:
    """Filter the list of matches based on the provided data type."""
    subset_match_list = []
    did_not_match_list = []
    # iterate through all of the matches
    for match in match_list:
        # if the current match is of the
        # specified type, then keep it in
        # the list of the matching matches
        if isinstance(match, data_type):
            subset_match_list.append(match)
        # if the current match is not of the
        # specified type, then keep it in
        # the list of non-matching matches
        else:
            did_not_match_list.append(match)
    # return both of the created lists
    return (subset_match_list, did_not_match_list)


def organize_matches(
    match_list: List[pyastgrepsearch.Match],
) -> Dict[str, List[pyastgrepsearch.Match]]:
    """Organize the matches on a per-file basis to support simplified processing."""
    match_dict: Dict[str, List[pyastgrepsearch.Match]] = {}
    # iterate through each of the matches in the list, with
    # the goal of creating a dictionary organized so that
    # --> the key is the name of a file under analysis
    # --> the value is a list of all of the matches for that file
    for current_match in match_list:
        # extract the name of the file for the current match
        current_match_file_name = str(current_match.path)
        # already storing matches for this file
        if current_match_file_name in match_dict:
            # extract the existing list of matches for this file
            current_match_file_list = match_dict[current_match_file_name]
            current_match_file_list.append(current_match)
        # not already storing matches for this file
        else:
            # create an empty list for storing the matches
            current_match_list: List[pyastgrepsearch.Match] = []
            # add the current match to the list
            current_match_list.append(current_match)
            # associate this new list with the current file name
            match_dict[current_match_file_name] = current_match_list
    return match_dict


def combine_dicts(dict_list: List[Dict[Any, Any]]) -> str:
    """Combine all dictionaries in the list into a single list of dictionaries as a string."""
    # combine all of the dictionaries in the list into
    # a single string that is a list of each JSON-based
    # dictionary represented as a string; this leads to:
    # [
    #   { nested JSON for file 1 },
    #   { nested JSON for file 2 },
    #   ...
    #   { nested JSON for file n }
    # ]
    # which is a list of valid JSON objects
    return json.dumps(dict_list, indent=2)



"""Support and configure debugging."""

from enum import Enum


class DebugLevel(str, Enum):
    """The predefined levels for debugging."""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class DebugDestination(str, Enum):
    """The destination for debugging."""

    CONSOLE = "CONSOLE"
    SYSLOG = "SYSLOG"



from pathlib import Path

import openai
from cryptography.fernet import Fernet

genscript = """
checks:
  - name: "A human-readable name for the check, describing its purpose or objective. For example, 'Class Definition' or 'Function Naming Convention'."
    code: "A short, unique identifier or code that distinguishes the check from others. This code is often used for reference or automation."
    id: "A unique identifier or tag that provides additional context to the check. It may include information such as the check's category or origin. For instance, 'C001' could represent a 'Code Quality' category."
    pattern: "An XPath expression specifying the exact code pattern or element that the check aims to locate within the source code. This expression serves as the search criteria for the check's evaluation."
    count:
      min: "The minimum number of times the specified pattern is expected to appear in the source code. It establishes the lower limit for the check's results, ensuring compliance or a minimum level of occurrence."
      max: "The maximum number of times the specified pattern is expected to appear in the source code. It defines an upper limit for the check's results, helping to identify potential issues or excessive occurrences."

Example:

  - name: "class-definition"
    code: "CDF"
    id: "C001"
    pattern: './/ClassDef'
    count:
      min: 1
      max: 50
  - name: "all-function-definition"
    code: "AFD"
    id: "F001"
    pattern: './/FunctionDef'
    count:
      min: 1
      max: 200
  - name: "non-test-function-definition"
    code: "NTF"
    id: "F002"
    pattern: './/FunctionDef[not(contains(@name, "test_"))]'
    count:
      min: 40
      max: 70
  - name: "single-nested-if"
    code: "SNI"
    id: "CL001"
    pattern: './/FunctionDef/body//If'
    count:
      min: 1
      max: 100
  - name: "double-nested-if"
    code: "DNI"
    id: "CL002"
    pattern: './/FunctionDef/body//If[ancestor::If and not(parent::orelse)]'
    count:
      min: 1
      max: 15

"""

API_KEY_FILE = "userapikey.txt"


def save_user_api_key(user_api_key):
    key = Fernet.generate_key()
    fernet = Fernet(key)
    encrypted_key = fernet.encrypt(user_api_key.encode()).decode()
    with open(API_KEY_FILE, "w") as f:
        f.write(key.decode() + "\n" + encrypted_key)


def load_user_api_key(file):
    with open(file, "r") as f:
        lines = f.read().strip().split("\n")
        if len(lines) == 2:  # noqa: PLR2004
            key = lines[0].encode()
            encrypted_key = lines[1]
        fernet = Fernet(key)
        return fernet.decrypt(encrypted_key.encode()).decode()


def is_valid_api_key(api_key):
    try:
        openai.api_key = api_key
        openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "system", "content": "Test message"}],
        )
        return True
    except openai.error.OpenAIError:
        return False


def generate_yaml_config(file: Path, user_api_key, user_input: str) -> str:
    try:
        openai.api_key = user_api_key

        prompts = [
            genscript
            + "in the same format as what is shown above(do not just generate the example use it as a framework nothing else): "
            + user_input
        ]

        response = openai.ChatCompletion.create(  # type: ignore
            model="gpt-3.5-turbo",
            messages=[
                {
                    "role": "system",
                    "content": f"You are a helpful assistant that generates YAML configurations. Your task is to {prompts}",
                }
            ],
        )

        generated_yaml = response.choices[0].message["content"].strip()
        file.touch()

        with open(file, "w") as f:
            f.write(generated_yaml)

        return generated_yaml

    except openai.error.OpenAIError:  # type: ignore
        return "[red][Error][/red] There was an issue with the API key. Make sure you input your API key correctly."



"""Mange the SQLite database containing results from chasten analyses."""

import subprocess
import sys
from pathlib import Path

from sqlite_utils import Database

from chasten import constants, enumerations, filesystem, output, util

CHASTEN_SQL_SELECT_QUERY = """
SELECT
  main.configuration_chastenversion as chastenversion,
  main.configuration_projectname as projectname,
  main.configuration_datetime as datetime,
  sources.filename,
  sources.check_id,
  sources.check_name,
  sources.check_description,
  sources.check_pattern,
  sources.check_min,
  sources.check_max,
  sources.check_passed,
  sources_check_matches.lineno,
  sources_check_matches.coloffset,
  sources_check_matches.linematch,
  sources_check_matches.linematch_context
FROM
  sources
  JOIN sources_check_matches ON sources._link = sources_check_matches._link_sources
  JOIN main ON sources._link_main = main._link
ORDER BY
  datetime desc;
"""

# create a small bullet for display in the output
small_bullet_unicode = constants.markers.Small_Bullet_Unicode


def create_chasten_view(chasten_database_name: str) -> None:
    """Create a view that combines results in the database tables."""
    database = Database(chasten_database_name)
    # create a "virtual table" (i.e., a view) that is the result
    # of running the pre-defined query; note that this query
    # organizes all of chasten's results into a single table.
    # When using datasette each of the columns in this view
    # are "facetable" which means that they can be enabled or disabled
    # inside of the web-based user interface
    database.create_view(
        constants.chasten.Chasten_Database_View, CHASTEN_SQL_SELECT_QUERY
    )


def enable_full_text_search(chasten_database_name: str) -> None:
    """Enable full-text search in the specific SQLite3 database."""
    database = Database(chasten_database_name)
    # enable full-text search on the main database table
    database["main"].enable_fts(
        [
            "configuration_chastenversion",
            "configuration_projectname",
            "configuration_datetime",
        ]
    )
    # enable full-text search on the sources database table
    database["sources"].enable_fts(
        [
            "filename",
            "check_id",
            "check_name",
            "check_description",
            "check_pattern",
        ]
    )
    # enable full-text search on the sources database table
    database["sources_check_matches"].enable_fts(
        [
            "lineno",
            "coloffset",
            "linematch",
        ]
    )
    # note that sqlite-utils does not support the enabling of
    # full-text search on the view called chasten_complete


def display_final_diagnostic_message(datasette_platform: str, publish: bool):
    """Output the final diagnostic message before control is given to a different tool."""
    # output a "final" prompt about either the publication platform of a reminder
    # that the remainder of the output comes from running a local datasette instance
    # the database will be published to an external platform
    if publish:
        output.console.print(
            f":sparkles: Debugging output from publishing datasette to '{datasette_platform}':"
        )
    # the database will be displayed through a localhost-based server
    else:
        output.console.print(
            ":sparkles: Debugging output from the local datasette server:"
        )
    output.console.print()


def display_datasette_details(
    label: str,
    virtual_env_location: str,
    executable_path: str,
    full_executable_name: str,
) -> None:
    """Display details about the current datasette configuration."""
    # output diagnostic information about the datasette instance; note
    # that the output must appear here and not from the calling function
    # because once the datasette instance starts the chasten tool can
    # no longer produce output in the console
    output.console.print()
    output.console.print(label)
    output.console.print(
        f"{constants.markers.Indent}{small_bullet_unicode} Venv: '{output.shorten_file_name(str(virtual_env_location), 120)}'"
    )
    if executable_path:
        output.console.print(
            f"{constants.markers.Indent}{small_bullet_unicode} Program: '{output.shorten_file_name(executable_path, 120)}'"
        )
    else:
        output.console.print(
            f"{constants.markers.Indent}{small_bullet_unicode} Cannot find: '{output.shorten_file_name(full_executable_name, 120)}'"
        )
    output.console.print()


def start_datasette_server(  # noqa: PLR0912, PLR0913
    database_path: Path,
    datasette_metadata: Path,
    datasette_platform: str = enumerations.DatasettePublicationPlatform.FLY.value,
    datasette_port: int = 8001,
    publish: bool = False,
    OpSystem: str = "Linux",
) -> None:
    """Start a local datasette server."""
    # define the name of the executable needed to run the server
    # define the name of the file that contains datasette metadata;
    # note that by default the metadata could be None and thus it
    # will not be passed as a -m argument to the datasette program
    metadata = datasette_metadata
    # identify the location at which the virtual environment exists;
    # note that this is the location where executable dependencies of
    # chasten will exist in a bin directory. For instance, the "datasette"
    # executable that is a dependency of chasten can be found by starting
    # the search from this location for the virtual environment.
    full_executable_name = util.executable_name(
        constants.datasette.Datasette_Executable, OpSystem
    )
    (found_executable, executable_path) = filesystem.can_find_executable(
        full_executable_name
    )
    # output diagnostic information about the datasette instance; note
    # that the output must appear here and not from the calling function
    # because once the datasette instance starts the chasten tool can
    # no longer produce output in the console
    if publish:
        label = ":sparkles: Details for datasette publishing:"
    else:
        label = ":sparkles: Details for datasette startup:"
    display_datasette_details(
        label,
        sys.prefix,
        str(executable_path),
        full_executable_name,
    )
    # since it was not possible to find the executable for datasette, display and
    # error message and then exit this function since no further steps are possible
    if not found_executable:
        output.console.print(
            f":person_shrugging: Was not able to find {constants.datasette.Datasette_Executable}"
        )
        return None
    # run the localhost server because the
    # function was not asked to publish a database
    if not publish:
        # the metadata parameter should not be passed to the datasette
        # program if it was not specified as an option
        if metadata is not None:
            cmd = [
                str(full_executable_name),
                str(database_path),
                "-m",
                str(metadata),
                "-p",
                str(datasette_port),
            ]
        else:
            cmd = [
                str(full_executable_name),
                str(database_path),
                "-p",
                str(datasette_port),
            ]
        # run the datasette server as a subprocess of chasten;
        # note that the only way to stop the server is to press CTRL-C;
        # there is debugging output in the console to indicate this option.
        proc = subprocess.Popen(cmd)
        proc.wait()
    # publish the datasette instance to the chosen datasette platform
    elif publish:
        # get information about the datasette executable, confirming that
        # it is available in the virtual environment created by chasten
        (
            found_publish_platform_executable,
            publish_platform_executable,
        ) = filesystem.can_find_executable(util.executable_name(datasette_platform))
        # was not able to find the fly or vercel executable (the person using this
        # program has to install separately, following the instructions for the
        # datasette-publish-fly plugin) and thus need to exit and not proceed
        if not found_publish_platform_executable:
            output.console.print(
                f":person_shrugging: Was not able to find '{datasette_platform}'"
            )
            return None
        # was able to find the fly or vercel executable that will support the
        # publication of this datasette instance to the platform
        else:
            output.console.print(
                f":sparkles: Using '{publish_platform_executable}' to publish a datasette"
            )
            output.console.print()
        display_final_diagnostic_message(datasette_platform, publish)
        # create the customized running argument for either fly or vercel; note
        # that these programs take different arguments for specifying the name
        # of the application as it will be deployed on the platform
        running_argument = ""
        if datasette_platform == constants.chasten.Executable_Fly:
            running_argument = "--app=chasten"
        elif datasette_platform == constants.chasten.Executable_Vercel:
            running_argument = "--project=chasten"
        # the metadata parameter should not be passed to the datasette
        # program if it was not specified as an option
        # there was a metadata parameter, so include it
        if metadata is not None:
            cmd = [
                str(full_executable_name),
                "publish",
                datasette_platform,
                str(database_path),
                running_argument,
                constants.datasette.Datasette_Copyable_Install,
                constants.datasette.Datasette_Export_Notebook,
                constants.datasette.Datasette_Search_All,
                "-m",
                str(metadata),
            ]
        # there was not a metadata parameter, so include it
        else:
            cmd = [
                str(full_executable_name),
                "publish",
                datasette_platform,
                str(database_path),
                running_argument,
                constants.datasette.Datasette_Copyable_Install,
                constants.datasette.Datasette_Export_Notebook,
                constants.datasette.Datasette_Search_All,
            ]
        # run the datasette server as a subprocess of chasten;
        # note that the only way to stop the server is to press CTRL-C;
        # there is debugging output in the console to indicate this option.
        proc = subprocess.Popen(cmd)
        proc.wait()


def display_results_frog_mouth(result_file, OpSystem) -> None:
    """Run frogmouth as a subprocess of chasten"""
    cmd = [
        "frogmouth",
        result_file,
    ]
    executable = util.executable_name("frogmouth", OpSystem)
    exec_found, executable_path = filesystem.can_find_executable(executable)
    if exec_found:
        # run frogmouth with specified path
        output.console.print("\n🐸 Frogmouth Information\n")
        output.console.print(f" {small_bullet_unicode} Venv: {sys.prefix}")
        output.console.print(f" {small_bullet_unicode} Program: {executable_path}")
        proc = subprocess.Popen(cmd)
        proc.wait()
    else:
        output.console.print(
            ":person_shrugging: Was not able to find frogmouth executable try installing it separately"
        )
        return None



"""💫 Chasten checks the AST of a Python program."""

import os
import sys
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union

import pyastgrep  # type: ignore
import typer
from pyastgrep import search as pyastgrepsearch  # type: ignore

from chasten import (
    checks,
    configApp,
    configuration,
    constants,
    createchecks,
    database,
    debug,
    enumerations,
    filesystem,
    output,
    process,
    results,
    server,
    util,
)

# create a Typer object to support the command-line interface
cli = typer.Typer(no_args_is_help=True)
app = configApp.config_App()
# create a small bullet for display in the output
small_bullet_unicode = constants.markers.Small_Bullet_Unicode
CHECK_STORAGE = constants.chasten.App_Storage
API_KEY_STORAGE = constants.chasten.API_Key_Storage
ANALYSIS_FILE = constants.chasten.Analyze_Storage


# ---
# Region: Helper functions {{{
# ---


def output_preamble(
    verbose: bool,
    debug_level: debug.DebugLevel = debug.DebugLevel.ERROR,
    debug_destination: debug.DebugDestination = debug.DebugDestination.CONSOLE,
    **kwargs,
) -> None:
    """Output all of the preamble content."""
    # setup the console and the logger through the output module
    output.setup(debug_level, debug_destination)
    output.logger.debug(f"Display verbose output? {verbose}")
    output.logger.debug(f"Debug level? {debug_level.value}")
    output.logger.debug(f"Debug destination? {debug_destination.value}")
    # display the header
    output.print_header()
    # display details about configuration as
    # long as verbose output was requested;
    # note that passing **kwargs to this function
    # will pass along all of the extra keyword
    # arguments that were input to the function
    output.print_diagnostics(
        verbose,
        debug_level=debug_level.value,
        debug_destination=debug_destination.value,
        **kwargs,
    )


def display_serve_or_publish_details(
    label: str,
    database_path: Path,
    metadata: Path,
    port: int = 8001,
    publish: bool = False,
) -> None:
    """Display diagnostic details at startup of serve or publish commands."""
    # output diagnostic information about the datasette instance
    output.console.print()
    output.console.print(label)
    output.console.print(
        f"{constants.markers.Indent}{small_bullet_unicode} Database: '{output.shorten_file_name(str(database_path), 120)}'"
    )
    output.console.print(
        f"{constants.markers.Indent}{small_bullet_unicode} Metadata: '{output.shorten_file_name(str(metadata), 120)}'"
    )
    # do not display a port if the task is publishing to fly.io
    # because that step does not support port specification
    if not publish:
        output.console.print(
            f"{constants.markers.Indent}{small_bullet_unicode} Port: {port}"
        )


# ---
# End region: Helper functions }}}
# ---

# ---
# Start region: Command-line interface functions {{{
# ---


@cli.command()
def create_checks(
    filename: Path = typer.Option("checks.yml", help="YAML file name")
) -> None:
    """🔧 Interactively specify for checks and have a checks.yml file created(Requires API key)"""
    # creates a textual object for better user interface
    app.run()
    # Checks if the file storing the wanted checks exists and is valid
    if filesystem.confirm_valid_file(CHECK_STORAGE):
        # stores the human readable version of the checks
        result = configApp.write_checks(configApp.split_file(CHECK_STORAGE))
        # Checks if API key storage file exists
        if filesystem.confirm_valid_file(API_KEY_STORAGE):
            # prints the human readable checks to the terminal
            output.console.print(result)
            # loads the decrypted API Key
            api_key = createchecks.load_user_api_key(API_KEY_STORAGE)
            # calls the function to generate the yaml file
            output.console.print(
                createchecks.generate_yaml_config(filename, api_key, result)
            )
        else:
            # prompts the user to input there API key to the terminal
            api_key = input("Please Enter your openai API Key:")
            # If not a valid API key prompts user again
            while not createchecks.is_valid_api_key(api_key):
                output.console.print(
                    "[red][ERROR][/red] Invalid API key. Please enter a valid API key."
                )
                api_key = input("Please Enter your openai API Key:")
            # stores the API key in a file
            createchecks.save_user_api_key(api_key)
            # prints the human readable checks to the terminal
            output.console.print(result)
            # gets the decrypted API Key
            api_key = createchecks.load_user_api_key(API_KEY_STORAGE)
            # prints the generated YAML file to the terminal
            output.console.print(
                createchecks.generate_yaml_config(filename, api_key, result)
            )
    else:
        # displays an error message if the CHECK_STORAGE file does not exist
        output.console.print(
            f"[red][ERROR][/red] No {CHECK_STORAGE} file exists\n  - Rerun the command and specify checks"
        )


@cli.command()
def configure(  # noqa: PLR0913
    task: enumerations.ConfigureTask = typer.Argument(
        enumerations.ConfigureTask.VALIDATE.value
    ),
    config: str = typer.Option(
        None,
        "--config",
        "-c",
        help="A directory with configuration file(s), path to configuration file, or URL to configuration file.",
    ),
    debug_level: debug.DebugLevel = typer.Option(
        debug.DebugLevel.ERROR.value,
        "--debug-level",
        "-l",
        help="Specify the level of debugging output.",
    ),
    debug_destination: debug.DebugDestination = typer.Option(
        debug.DebugDestination.CONSOLE.value,
        "--debug-dest",
        "-t",
        help="Specify the destination for debugging output.",
    ),
    force: bool = typer.Option(
        False,
        help="Create configuration directory and files even if they exist",
    ),
    verbose: bool = typer.Option(False, help="Display verbose debugging output"),
) -> None:
    """🪂 Manage chasten's configuration."""
    # output the preamble, including extra parameters specific to this function
    output_preamble(
        verbose,
        debug_level,
        debug_destination,
        task=task.value,
        config=config,
        force=force,
    )
    # setup the console and the logger through the output module
    output.setup(debug_level, debug_destination)
    output.logger.debug(f"Display verbose output? {verbose}")
    output.logger.debug(f"Debug level? {debug_level.value}")
    output.logger.debug(f"Debug destination? {debug_destination.value}")
    # display the configuration directory and its contents
    if task == enumerations.ConfigureTask.VALIDATE:
        # validate the configuration files:
        # --> config.yml (or url pointing to one)
        # --> checks.yml (or whatever file/url is reference in config.yml)
        (validated, _) = configuration.validate_configuration_files(config, verbose)
        # some aspect of the configuration was not
        # valid, so exit early and signal an error
        if not validated:
            output.console.print(
                "\n:person_shrugging: Cannot perform analysis due to configuration error(s).\n"
            )
            sys.exit(constants.markers.Non_Zero_Exit)
    # create the configuration directory and a starting version of the configuration file
    if task == enumerations.ConfigureTask.CREATE:
        # attempt to create the configuration directory
        try:
            # create the configuration directory, which will either be the one
            # specified by the config parameter (if it exists) or it will be
            # the one in the platform-specific directory given by platformdirs
            if config is None:
                configuration_directory = None
            else:
                configuration_directory = Path(config)
            created_directory_path = filesystem.create_configuration_directory(
                configuration_directory, force
            )
            # write the configuration file for the chasten tool in the created directory
            filesystem.create_configuration_file(
                created_directory_path,
                constants.filesystem.Main_Configuration_File,
            )
            # write the check file for the chasten tool in the created directory
            filesystem.create_configuration_file(
                created_directory_path, constants.filesystem.Main_Checks_File
            )
            # display diagnostic information about the completed process
            output.console.print(
                f":sparkles: Created configuration directory and file(s) in {created_directory_path}"
            )
        # cannot re-create the configuration directory, so display
        # a message and suggest the use of --force the next time;
        # exit early and signal an error with a non-zero exist code
        except FileExistsError:
            if not force:
                output.console.print(
                    "\n:person_shrugging: Configuration directory already exists."
                )
                output.console.print(
                    "Use --force to recreate configuration directory and its containing files."
                )
            sys.exit(constants.markers.Non_Zero_Exit)


@cli.command()
def analyze(  # noqa: PLR0912, PLR0913, PLR0915
    project: str = typer.Argument(help="Name of the project."),
    xpath: Path = typer.Option(
        str,
        "--xpath-version",
        "-xp",
        help="Accepts different xpath version, runs xpath version two by default.",
    ),
    check_include: Tuple[enumerations.FilterableAttribute, str, int] = typer.Option(
        (None, None, 0),
        "--check-include",
        "-i",
        help="Attribute name, value, and match confidence level for inclusion.",
    ),
    check_exclude: Tuple[enumerations.FilterableAttribute, str, int] = typer.Option(
        (None, None, 0),
        "--check-exclude",
        "-e",
        help="Attribute name, value, and match confidence level for exclusion.",
    ),
    input_path: Path = typer.Option(
        filesystem.get_default_directory_list(),
        "--search-path",
        "-d",
        help="A path (i.e., directory or file) with Python source code(s).",
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    ),
    output_directory: Path = typer.Option(
        None,
        "--save-directory",
        "-s",
        help="A directory for saving output file(s).",
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    view_XML: str = typer.Option(
        None,
        "--view-xml",
        "-v",
        help="Prints and saves the xml representation of the input file(s)",
    ),
    save_XML: str = typer.Option(
        None,
        "--save-xml",
        "-sx",
        help="Saves the xml representation of the input file(s)",
    ),
    store_result: Path = typer.Option(
        None,
        "--markdown-storage",
        "-r",
        help="A directory for storing results in a markdown file",
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    config: str = typer.Option(
        None,
        "--config",
        "-c",
        help="A directory with configuration file(s) or URL to configuration file.",
    ),
    debug_level: debug.DebugLevel = typer.Option(
        debug.DebugLevel.ERROR.value,
        "--debug-level",
        "-l",
        help="Specify the level of debugging output.",
    ),
    debug_destination: debug.DebugDestination = typer.Option(
        debug.DebugDestination.CONSOLE.value,
        "--debug-dest",
        "-t",
        help="Specify the destination for debugging output.",
    ),
    display: bool = typer.Option(False, help="Display results using frogmouth"),
    verbose: bool = typer.Option(False, help="Enable verbose mode output."),
    save: bool = typer.Option(False, help="Enable saving of output file(s)."),
    force: bool = typer.Option(False, help="Force creation of new markdown file"),
) -> None:
    """💫 Analyze the AST of Python source code."""
    # setup the console and the logger through the output module
    output.setup(debug_level, debug_destination)
    output.logger.debug(f"Display verbose output? {verbose}")
    output.logger.debug(f"Debug level? {debug_level.value}")
    output.logger.debug(f"Debug destination? {debug_destination.value}")
    start_time = time.time()
    output.logger.debug("Analysis Started.")
    # output the preamble, including extra parameters specific to this function
    output_preamble(
        verbose,
        debug_level,
        debug_destination,
        project=project,
        directory=input_path,
    )
    # extract the current version of the program
    chasten_version = util.get_chasten_version()
    # display current chasten version
    output.logger.debug(f"Current version of chasten: {chasten_version}")
    # create the include and exclude criteria
    include = results.CheckCriterion(
        attribute=str(checks.fix_check_criterion(check_include[0])),
        value=str(checks.fix_check_criterion(check_include[1])),
        confidence=int(checks.fix_check_criterion(check_include[2])),
    )
    exclude = results.CheckCriterion(
        attribute=str(checks.fix_check_criterion(check_exclude[0])),
        value=str(checks.fix_check_criterion(check_exclude[1])),
        confidence=int(checks.fix_check_criterion(check_exclude[2])),
    )
    # create a configuration that is the same for all results
    chasten_configuration = results.Configuration(
        chastenversion=chasten_version,
        projectname=project,
        configdirectory=Path(config),
        searchpath=input_path,
        debuglevel=debug_level,
        debugdestination=debug_destination,
        checkinclude=include,
        checkexclude=exclude,
    )
    # connect the configuration to the top-level chasten object for results saving
    # note: this is the final object that contains all of the data
    chasten_results_save = results.Chasten(configuration=chasten_configuration)
    # add extra space after the command to run the program
    output.console.print()
    # validate the configuration
    (validated, checks_dict) = configuration.validate_configuration_files(
        config, verbose
    )
    # some aspect of the configuration was not
    # valid, so exit early and signal an error
    if not validated:
        output.console.print(
            "\n:person_shrugging: Cannot perform analysis due to configuration error(s).\n"
        )
        output.logger.debug("Cannot perform analysis due to configuration error(s)")
        sys.exit(constants.markers.Non_Zero_Exit)
    # extract the list of the specific patterns (i.e., the XPATH expressions)
    # that will be used to analyze all of the XML-based representations of
    # the Python source code found in the valid directories
    check_list: List[Dict[str, Union[str, Dict[str, int]]]] = checks_dict[
        constants.checks.Checks_Label
    ]
    # filter the list of checks based on the include and exclude parameters
    # --> only run those checks that were included
    check_list = process.include_or_exclude_checks(  # type: ignore
        check_list, include=True, *check_include
    )
    # --> remove those checks that were excluded
    check_list = process.include_or_exclude_checks(  # type: ignore
        check_list, include=False, *check_exclude
    )
    # the specified search path is not valid and thus it is
    # not possible to analyze the Python source files in this directory
    # OR
    # the specified search path is not valid and thus it is
    # not possible to analyze the specific Python source code file
    if not filesystem.confirm_valid_directory(
        input_path
    ) and not filesystem.confirm_valid_file(input_path):
        output.console.print(
            "\n:person_shrugging: Cannot perform analysis due to invalid search directory.\n"
        )
        sys.exit(constants.markers.Non_Zero_Exit)
    if store_result:
        # creates an empty string for storing results temporarily
        analysis_result = ""
        analysis_file_dir = store_result / ANALYSIS_FILE
        # clears markdown file of results if it exists and new results are to be store
        if filesystem.confirm_valid_file(analysis_file_dir):
            if not force:
                if display:
                    database.display_results_frog_mouth(
                        analysis_file_dir, util.get_OS()
                    )
                    sys.exit(0)
                else:
                    output.console.print(
                        "File already exists: use --force to recreate markdown directory."
                    )
                    sys.exit(constants.markers.Non_Zero_Exit)
            else:
                analysis_file_dir.write_text("")
        # creates file if doesn't exist already
        analysis_file_dir.touch()
    # create the list of directories
    valid_directories = [input_path]
    # output the list of directories subject to checking
    output.console.print()
    output.console.print(f":sparkles: Analyzing Python source code in: {input_path}")
    # output the number of checks that will be performed
    output.console.print()
    output.console.print(f":tada: Performing {len(check_list)} check(s):")
    output.console.print()
    # create a check_status list for all of the checks
    check_status_list: List[bool] = []
    # check XPATH version
    if xpath == "1.0":
        output.logger.debug("Using XPath version 1.0")
    else:
        output.logger.debug("Using XPath version 2.0")
    # iterate through and perform each of the checks
    for current_check in check_list:
        # extract the pattern for the current check
        current_xpath_pattern = str(
            current_check[constants.checks.Check_Pattern]
        )  # type: ignore
        # extract the minimum and maximum values for the checks, if they exist
        # note that this function will return None for a min or a max if
        # that attribute does not exist inside of the current_check; importantly,
        # having a count or a min or a max is all optional in a checks file
        (min_count, max_count) = checks.extract_min_max(current_check)
        # extract details about the check to display in the header
        # of the syntax box for this specific check
        check_id = current_check[constants.checks.Check_Id]  # type: ignore
        output.logger.debug(f"check id: {check_id}")
        check_name = current_check[constants.checks.Check_Name]  # type: ignore
        check_description = checks.extract_description(current_check)
        # search for the XML contents of an AST that match the provided
        # XPATH query using the search_python_file in search module of pyastgrep;
        # this looks for matches across all path(s) in the specified source path
        # match_generator = pyastgrepsearch.search_python_files(
        #         paths=valid_directories, expression=current_xpath_pattern, xpath2=True
        # )
        if xpath == "1.0":
            match_generator = pyastgrepsearch.search_python_files(
                paths=valid_directories, expression=current_xpath_pattern, xpath2=False
            )
        else:
            match_generator = pyastgrepsearch.search_python_files(
                paths=valid_directories, expression=current_xpath_pattern, xpath2=True
            )
        # materia>>> mastlize a list from the generator of (potential) matches;
        # note that this list will also contain an object that will
        # indicate that the analysis completed for each located file
        match_generator_list = list(match_generator)
        # filter the list of matches so that it only includes
        # those that are a Match object that will contain source code
        (match_generator_list, _) = process.filter_matches(
            match_generator_list, pyastgrepsearch.Match
        )
        # organize the matches according to the file to which they
        # correspond so that processing of matches takes place per-file
        match_dict = process.organize_matches(match_generator_list)
        # perform an enforceable check if it is warranted for this check
        current_check_save = None
        if checks.is_checkable(min_count, max_count):
            # determine whether or not the number of found matches is within mix and max
            check_status = checks.check_match_count(
                len(match_generator_list), min_count, max_count
            )
            # keep track of the outcome for this check
            check_status_list.append(check_status)
        # this is not an enforceable check and thus the tool always
        # records that the checked passed as a default
        else:
            check_status = True
        # convert the status of the check to a visible symbol for display
        check_status_symbol = util.get_symbol_boolean(check_status)
        # escape the open bracket symbol that may be in an XPATH expression
        # and will prevent it from displaying correctly
        current_xpath_pattern_escape = current_xpath_pattern.replace("[", "\\[")
        # display minimal diagnostic output
        output.console.print(
            f"  {check_status_symbol} id: '{check_id}', name: '{check_name}'"
            + f", pattern: '{current_xpath_pattern_escape}', min={min_count}, max={max_count}"
        )
        if store_result:
            # makes the check marks or x's appear as words instead for markdown
            check_pass = (
                "PASSED:"
                if check_status_symbol == "[green]\u2713[/green]"
                else "FAILED:"
            )
            # stores check type in a string to stored in file later
            analysis_result += (
                f"\n# {check_pass} **ID:** '{check_id}', **Name:** '{check_name}'"
                + f", **Pattern:** '{current_xpath_pattern_escape}', min={min_count}, max={max_count}\n\n"
            )

        # for each potential match, log and, if verbose model is enabled,
        # display details about each of the matches
        current_result_source = results.Source(
            filename=str(str(vd) for vd in valid_directories)
        )
        # there were no matches and thus the current_check_save of None
        # should be recorded inside of the source of the results
        if len(match_generator_list) == 0:
            current_result_source.check = current_check_save
        # iteratively analyze:
        # a) A specific file name
        # b) All of the matches for that file name
        # Note: the goal is to only process matches for a
        # specific file, ensuring that matches for different files
        # are not mixed together, which would contaminate the results
        # Note: this is needed because using pyastgrepsearch will
        # return results for all of the files that matched the check
        for file_name, matches_list in match_dict.items():
            # create the current check
            current_check_save = results.Check(
                id=check_id,  # type: ignore
                name=check_name,  # type: ignore
                description=check_description,  # type: ignore
                min=min_count,  # type: ignore
                max=max_count,  # type: ignore
                pattern=current_xpath_pattern,
                passed=check_status,
            )
            # create a source that is solely for this file name
            current_result_source = results.Source(filename=file_name)
            # put the current check into the list of checks in the current source
            current_result_source.check = current_check_save
            # display minimal diagnostic output
            output.console.print(
                f"    {small_bullet_unicode} {file_name} - {len(matches_list)} matches"
            )
            if store_result:
                # stores details of checks in string to be stored later
                analysis_result += f"    - {file_name} - {len(matches_list)} matches\n"
            # extract the lines of source code for this file; note that all of
            # these matches are organized for the same file and thus it is
            # acceptable to extract the lines of the file from the first match
            # a long as there are matches available for analysis
            if len(matches_list) > 0:
                current_result_source._filelines = matches_list[0].file_lines
            # iterate through all of the matches that are specifically
            # connected to this source that is connected to a specific file name
            for current_match in matches_list:
                if isinstance(current_match, pyastgrepsearch.Match):
                    current_result_source._filelines = current_match.file_lines
                    # extract the direct line number for this match
                    position_end = current_match.position.lineno
                    # extract the column offset for this match
                    column_offset = current_match.position.col_offset
                    # create a match specifically for this file;
                    # note that the AST starts line numbering at 1 and
                    # this means that storing the matching line requires
                    # the indexing of file_lines with position_end - 1;
                    # note also that linematch is the result of using
                    # lstrip to remove any blank spaces before the code
                    current_match_for_current_check_save = results.Match(
                        lineno=position_end,
                        coloffset=column_offset,
                        linematch=current_match.file_lines[position_end - 1].lstrip(
                            constants.markers.Space
                        ),
                        linematch_context=util.join_and_preserve(
                            current_match.file_lines,
                            max(
                                0,
                                position_end - constants.markers.Code_Context,
                            ),
                            position_end + constants.markers.Code_Context,
                        ),
                    )
                    # save the entire current_match that is an instance of
                    # pyastgrepsearch.Match for verbose debugging output as needed
                    current_check_save._matches.append(current_match)
                    # add the match to the listing of matches for the current check
                    current_check_save.matches.append(
                        current_match_for_current_check_save
                    )  # type: ignore
            # add the current source to main object that contains a list of source
            chasten_results_save.sources.append(current_result_source)
        # add the amount of total matches in each check to the end of each checks output
        output.console.print(f"   = {len(match_generator_list)} total matches\n")
    # calculate the final count of matches found
    total_result = util.total_amount_passed(check_status_list)
    # display checks passed, total amount of checks, and percentage of checks passed
    output.console.print(
        f":computer: {total_result[0]} / {total_result[1]} checks passed ({total_result[2]}%)\n"
    )
    # display all of the analysis results if verbose output is requested
    output.print_analysis_details(chasten_results_save, verbose=verbose)
    # save all of the results from this analysis
    saved_file_name = filesystem.write_chasten_results(
        output_directory, project, chasten_results_save, save
    )
    # output the name of the saved file if saving successfully took place
    if saved_file_name:
        output.console.print(f"\n:sparkles: Saved the file '{saved_file_name}'")
    # --save-xml and --view-xml
    if save_XML is not None or view_XML is not None:
        output.console.print(":memo: Saving XML...")
        try:
            if os.path.isdir(input_path):
                for each_file in os.listdir(input_path):
                    each_file = Path(input_path) / Path(each_file)  # type: ignore # noqa: PLW2901
                    if (
                        not os.path.isdir(each_file)
                        and os.path.isfile(each_file)
                        and str(each_file).endswith(".py")
                    ):
                        # Read the bytes of the input path and store them in the 'contents' variable
                        contents = Path(each_file).read_bytes()
                        # Use pyastgrep to parse the contents of the Python file at 'input_path'
                        _, ast = pyastgrep.files.parse_python_file(
                            contents, each_file, auto_dedent=False
                        )
                        # Convert the Abstract Syntax Tree (AST) into an XML representation
                        xml_root = pyastgrep.asts.ast_to_xml(ast, {})
                        # Check if view_xml is chosen
                        if view_XML is not None:
                            output.console.print(
                                pyastgrep.xml.tostring(
                                    xml_root, pretty_print=True
                                ).decode("utf-8")
                            )
                    elif os.path.isdir(each_file):
                        for sub_file in os.listdir(each_file):
                            sub_file = Path(each_file) / Path(sub_file)  # type: ignore # noqa: PLW2901
                            if str(sub_file).endswith(".py"):
                                contents = Path(sub_file).read_bytes()
                                _, ast = pyastgrep.files.parse_python_file(
                                    contents, sub_file, auto_dedent=False
                                )
                                xml_root = pyastgrep.asts.ast_to_xml(ast, {})
                                # Check if view_xml is chosen
                                if view_XML is not None:
                                    output.console.print(
                                        pyastgrep.xml.tostring(
                                            xml_root, pretty_print=True
                                        ).decode("utf-8")
                                    )
            elif os.path.isfile(input_path) and str(input_path).endswith(".py"):
                contents = Path(input_path).read_bytes()
                _, ast = pyastgrep.files.parse_python_file(
                    contents, input_path, auto_dedent=False
                )
                xml_root = pyastgrep.asts.ast_to_xml(ast, {})
                # Check if view_xml is chosen
                if view_XML is not None:
                    output.console.print(
                        pyastgrep.xml.tostring(xml_root, pretty_print=True).decode(
                            "utf-8"
                        )
                    )
        except FileNotFoundError:
            output.console.print(":sweat: Sorry, could not convert to xml.")
    # confirm whether or not all of the checks passed
    # and then display the appropriate diagnostic message
    all_checks_passed = all(check_status_list)
    end_time = time.time()
    elapsed_time = end_time - start_time

    if not all_checks_passed:
        output.console.print(":sweat: At least one check did not pass.")
        if store_result:
            # writes results of analyze into a markdown file
            analysis_file_dir.write_text(analysis_result, encoding="utf-8")
            output.console.print(
                f"\n:sparkles: Results saved in: {os.path.abspath(analysis_file_dir)}\n"
            )
        sys.exit(constants.markers.Non_Zero_Exit)
    output.console.print(
        f"\n:joy: All checks passed. Elapsed Time: {elapsed_time} seconds"
    )
    output.logger.debug("Analysis complete.")
    if store_result:
        # writes results of analyze into a markdown file
        result_path = os.path.abspath(analysis_file_dir)
        analysis_file_dir.write_text(analysis_result, encoding="utf-8")
        output.console.print(f"\n:sparkles: Results saved in: {result_path}\n")
        if display:
            database.display_results_frog_mouth(result_path, util.get_OS())


@cli.command()
def integrate(  # noqa: PLR0913
    project: str = typer.Argument(help="Name of the project."),
    json_path: List[Path] = typer.Argument(
        help="Directories, files, or globs for chasten's JSON result file(s).",
    ),
    output_directory: Path = typer.Option(
        ...,
        "--save-directory",
        "-s",
        help="A directory for saving converted file(s).",
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    debug_level: debug.DebugLevel = typer.Option(
        debug.DebugLevel.ERROR.value,
        "--debug-level",
        "-l",
        help="Specify the level of debugging output.",
    ),
    debug_destination: debug.DebugDestination = typer.Option(
        debug.DebugDestination.CONSOLE.value,
        "--debug-dest",
        "-t",
        help="Specify the destination for debugging output.",
    ),
    force: bool = typer.Option(
        False,
        help="Create converted results files even if they exist",
    ),
    verbose: bool = typer.Option(False, help="Display verbose debugging output"),
) -> None:
    """🚧 Integrate files and make a database."""
    # output the preamble, including extra parameters specific to this function
    output_preamble(
        verbose,
        debug_level,
        debug_destination,
        project=project,
        output_directory=output_directory,
        json_path=json_path,
        force=force,
    )
    output.logger.debug("Integrate function started.")
    # output the list of directories subject to checking
    output.console.print()
    output.console.print(":sparkles: Combining data file(s) in:")
    output.logger.debug(":sparkles: Combining data file(s) in:")
    output.console.print()
    output.print_list_contents(json_path)
    # extract all of the JSON dictionaries from the specified files
    json_dicts = filesystem.get_json_results(json_path)
    count = len(json_path)
    output.console.print(f"\n:sparkles: Total of {count} files in all directories.")
    # combine all of the dictionaries into a single string
    combined_json_dict = process.combine_dicts(json_dicts)
    # write the combined JSON file string to the filesystem
    combined_json_file_name = filesystem.write_dict_results(
        combined_json_dict, output_directory, project
    )
    # output the name of the saved file if saving successfully took place
    if combined_json_file_name:
        output.console.print(f"\n:sparkles: Saved the file '{combined_json_file_name}'")
        output.logger.debug(f"Saved the file '{combined_json_file_name}'.")
    # "flatten" (i.e., "un-nest") the now-saved combined JSON file using flatterer
    # create the SQLite3 database and then configure the database for use in datasette
    combined_flattened_directory = filesystem.write_flattened_csv_and_database(
        combined_json_file_name,
        output_directory,
        project,
    )
    output.logger.debug("Flattened JSON and created SQLite database.")
    # output the name of the saved file if saving successfully took place
    if combined_flattened_directory:
        output.console.print(
            f"\n:sparkles: Created this directory structure in {Path(combined_flattened_directory).parent}:"
        )
        combined_directory_tree = filesystem.create_directory_tree_visualization(
            Path(combined_flattened_directory)
        )
        output.console.print()
        output.console.print(combined_directory_tree)
        output.logger.debug("Integrate function completed successfully.")


@cli.command()
def datasette_serve(  # noqa: PLR0913
    database_path: Path = typer.Argument(
        help="SQLite3 database file storing chasten's results.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    port: int = typer.Option(
        8001,
        "--port",
        "-p",
        help="Port on which to run a datasette instance",
    ),
    metadata: Path = typer.Option(
        None,
        "--metadata",
        "-m",
        help="Meta-data file storing database configuration.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    debug_level: debug.DebugLevel = typer.Option(
        debug.DebugLevel.ERROR.value,
        "--debug-level",
        "-l",
        help="Specify the level of debugging output.",
    ),
    debug_destination: debug.DebugDestination = typer.Option(
        debug.DebugDestination.CONSOLE.value,
        "--debug-dest",
        "-t",
        help="Specify the destination for debugging output.",
    ),
    verbose: bool = typer.Option(False, help="Display verbose debugging output"),
) -> None:
    """🏃 Start a local datasette server."""
    # output the preamble, including extra parameters specific to this function
    output_preamble(
        verbose,
        debug_level,
        debug_destination,
        database=database_path,
        datasette_port=port,
        metadata=metadata,
    )
    # setup the console and the logger through the output module
    output.setup(debug_level, debug_destination)
    output.logger.debug(f"Display verbose output? {verbose}")
    output.logger.debug(f"Debug level? {debug_level.value}")
    output.logger.debug(f"Debug destination? {debug_destination.value}")
    # display diagnostic information about the datasette instance
    label = ":sparkles: Starting a local datasette instance:"
    display_serve_or_publish_details(
        label, database_path, metadata, port, publish=False
    )
    # start the datasette server that will run indefinitely;
    # shutting down the datasette server with a CTRL-C will
    # also shut down this command in chasten
    database.start_datasette_server(
        database_path=database_path,
        datasette_port=port,
        datasette_metadata=metadata,
        publish=False,
        OpSystem=util.get_OS(),
    )


@cli.command()
def datasette_publish(  # noqa: PLR0913
    database_path: Path = typer.Argument(
        help="SQLite3 database file storing chasten's results.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    metadata: Path = typer.Option(
        None,
        "--metadata",
        "-m",
        help="Meta-data file storing database configuration.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
    datasette_platform: enumerations.DatasettePublicationPlatform = typer.Option(
        enumerations.DatasettePublicationPlatform.FLY.value,
        "--platform",
        "-p",
        help="Specify the deployment platform for datasette.",
    ),
    debug_level: debug.DebugLevel = typer.Option(
        debug.DebugLevel.ERROR.value,
        "--debug-level",
        "-l",
        help="Specify the level of debugging output.",
    ),
    debug_destination: debug.DebugDestination = typer.Option(
        debug.DebugDestination.CONSOLE.value,
        "--debug-dest",
        "-t",
        help="Specify the destination for debugging output.",
    ),
    verbose: bool = typer.Option(False, help="Display verbose debugging output"),
) -> None:
    """🌎 Publish a datasette to Fly or Vercel."""
    # output the preamble, including extra parameters specific to this function
    output_preamble(
        verbose,
        debug_level,
        debug_destination,
        database=database_path,
        metadata=metadata,
    )
    # setup the console and the logger through the output module
    output.setup(debug_level, debug_destination)
    output.logger.debug(f"Display verbose output? {verbose}")
    output.logger.debug(f"Debug level? {debug_level.value}")
    output.logger.debug(f"Debug destination? {debug_destination.value}")
    output.console.print()
    output.console.print(
        f":wave: Make sure that you have previously logged into the '{datasette_platform.value}' platform"
    )
    # display details about the publishing step
    label = f":sparkles: Publishing a datasette to {datasette_platform.value}:"
    display_serve_or_publish_details(label, database_path, metadata, publish=True)
    # publish the datasette instance using fly.io;
    # this passes control to datasette and then to
    # the fly program that must be installed
    database.start_datasette_server(
        database_path=database_path,
        datasette_metadata=metadata,
        datasette_platform=datasette_platform.value,
        publish=True,
        OpSystem=util.get_OS(),
    )


@cli.command()
def log() -> None:
    """🦚 Start the logging server."""
    # display the header
    output.print_header()
    # display details about the server
    output.print_server()
    # run the server; note that this
    # syslog server receives debugging
    # information from chasten.
    # It must be started in a separate process
    # before running any sub-command
    # of the chasten tool
    server.start_syslog_server()


@cli.command()
def version():
    """🖥️  Display the version of Chasten."""
    # Get Chasten version from util file
    version_string = util.get_chasten_version()
    # output chasten version
    typer.echo(f"chasten {version_string}")


# ---
# End region: Command-line interface functions }}}
# ---



# Import necessary modules and components from the Textual library,
# as well as other Python modules like os and validation tools.
from pathlib import Path
from typing import ClassVar, List

from textual.app import App, ComposeResult
from textual.validation import Number
from textual.widgets import Button, Input, Pretty, Static

from chasten import constants

CHECK_STORAGE = constants.chasten.App_Storage
# Constants to map input field names to their positions in the Check list
CHECK_VALUE = {
    "Check": 0,
    "Matches": 1,
}
CHECK_DEFAULT = ["", "1", False]


def split_file(file_name: Path) -> List[List[str]]:
    """Split a csv file into a list of lists."""
    check_list = []
    with open(file_name) as file:
        for row in file:
            strip_row = row.strip()  # Remove leading/trailing white spaces
            if strip_row:
                check_list.append(strip_row.split(","))
    return check_list


def write_checks(check_list: List[List[str]]) -> str:
    """Generate structured output based on the contents of the file."""
    if len(check_list) != 0:
        result = "Make a YAML file that checks for:"
        for checks in check_list:
            quantity = "exactly" if checks[2] == "True" else "at minimum"
            result += f"\n - {quantity} {checks[1]} {checks[0]}"
        return result
    return "[red][ERROR][/red] No checks were supplied"


def store_in_file(File: Path, Pattern, Matches, Exact):
    """Store inputed values into a text file"""
    File.touch()
    with open(File, "a") as file:
        file.write(f"\n{Pattern},{Matches},{Exact}")  # Append input data to the file


# Define input fields and buttons for the user interface
Check_Input = Input(placeholder="Check For:", id="Check", name="Check")
Match_Input = Input(
    placeholder="How many matches do you expect",
    id="Matches",
    name="Matches",
    validators=Number(1, 500),  # Validate that Matches is a number between 1 and 500
)
Exact_button = Button("Exact", id="Exact")  # Button to trigger an action


# Static widget to display user input and validation results
class answers(Static):
    def compose(self) -> ComposeResult:
        """For displaying the user interface"""
        yield Check_Input
        yield Match_Input


# Static widget to display buttons for user interactions
class button_prompts(Static):
    def compose(self) -> ComposeResult:
        """For displaying the user interface"""
        yield Pretty([])  # Widget to display validation messages
        yield Exact_button  # Display the "Exact" button
        yield Button("Submit Check!", id="next")  # Display the "Next Check!" button
        yield Button("Done", id="done")
        yield Button("Clear Checks", id="clear", variant="error")


# Custom App class for the Textual application
class config_App(App):
    CSS = """
    Screen {
        layout: horizontal;
    }
    answers {
        width: 100%;
        align: center top;
        dock: top;
        background: $boost;
        min-width: 50;
        padding: 1;
        border: wide black;
    }
    button_prompts {
        background: $boost;
        layout: vertical;
        margin: 1;
        align: left top;
        border: wide black;
        width: 100%;
    }
    Button {
        background: rgb(245, 184, 50);
        content-align: center top;
        height: 3;
        width: 100%;
    }
    Button:hover {
        background: white;
        color: black;
    }
    """
    Check: ClassVar[list] = ["", "1", False]
    Valid: bool = False

    def on_input_changed(self, event: Input.Changed) -> None:
        """When inputs change this updates the values of Check"""
        self.Valid = False
        if event.input.id == "Check":
            self.Check[CHECK_VALUE[str(event.input.name)]] = event.input.value
        elif event.validation_result is not None:
            if event.validation_result.is_valid:
                self.Check[CHECK_VALUE[str(event.input.name)]] = event.input.value
                self.Valid = True

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "Exact":
            self.Check[2] = True  # Mark the "Exact" button as clicked
            event.button.disabled = True  # Disable the "Exact" button after clicking
        elif event.button.id == "done":
            config_App.exit(
                self
            )  # Exit the application if the "Done" button is clicked
        elif event.button.id == "clear":
            with open(CHECK_STORAGE, "w") as file:
                file.write(
                    ""
                )  # Clears Checks.txt file when "Clear Check" button is clicked
        elif self.Valid:
            if event.button.id == "next":
                # If "Next Check!" is clicked and input is valid, record the input data to a file
                store_in_file(
                    CHECK_STORAGE, self.Check[0], self.Check[1], self.Check[2]
                )
                self.Check[0] = ""
                self.Check[1] = "1"
                self.Check[2] = False
                # Reset input fields, clear validation messages, and enable the "Exact" button
                self.query_one(Pretty).update([])  # Clear any validation messages
                Exact_button.disabled = False  # Re-enable the "Exact" button
                Check_Input.value = ""
                Match_Input.value = ""  # Refresh the application UI
        else:
            self.query_one(Pretty).update(["Invalid Input Please enter a Integer"])
            Match_Input.value = ""  # Clear the "Matches" input field

    def compose(self) -> ComposeResult:
        """For displaying the user interface"""
        yield answers()  # Display the input fields for user input
        yield button_prompts()  # Display the buttons for user interaction



"""Utilities for use within chasten."""

import importlib.metadata
import platform
import sys

from urllib3.util import parse_url

from chasten import constants

checkmark_unicode = "\u2713"
xmark_unicode = "\u2717"
default_chasten_semver = "0.0.0"


def get_human_readable_boolean(answer: bool) -> str:
    """Produce a human-readable Yes or No for a boolean value of True or False."""
    # the provided answer is true
    if answer:
        return constants.humanreadable.Yes
    # the provided answer is false
    return constants.humanreadable.No


def get_OS() -> str:
    """Gets the Operating system of the user."""
    OpSystem = platform.system()
    return OpSystem


def executable_name(executable_name: str, OpSystem: str = "Linux") -> str:
    """Get the executable directory depending on OS"""
    exe_directory = "/bin/"
    # Checks if the OS is windows and changed where to search if true
    if OpSystem == "Windows":
        exe_directory = "/Scripts/"
        executable_name += ".exe"
    virtual_env_location = sys.prefix
    return virtual_env_location + exe_directory + executable_name


def get_symbol_boolean(answer: bool) -> str:
    """Produce a symbol-formatted version of a boolean value of True or False."""
    if answer:
        return f"[green]{checkmark_unicode}[/green]"
    return f"[red]{xmark_unicode}[/red]"


def get_chasten_version() -> str:
    """Use importlib to extract the version of the package."""
    # attempt to determine the current version of the entire package,
    # bearing in mind that this program appears on PyPI with the name "chasten";
    # this will then return the version string specified with the version attribute
    # in the [tool.poetry] section of the pyproject.toml file
    try:
        version_string_of_foo = importlib.metadata.version(
            constants.chasten.Application_Name
        )
    # note that using the version function does not work when chasten is run
    # through a 'poetry shell' and/or a 'poetry run' command because at that stage
    # there is not a working package that importlib.metadata can access with a version;
    # in this situation the function should return the default value of 0.0.0
    except importlib.metadata.PackageNotFoundError:
        version_string_of_foo = default_chasten_semver
    return version_string_of_foo


def join_and_preserve(data, start, end):
    """Join and preserve lines inside of a list."""
    return constants.markers.Newline.join(data[start:end])


def is_url(url: str) -> bool:
    """Determine if string is valid URL."""
    # parse input url
    url_parsed = parse_url(url)
    # only allow http and https
    if url_parsed.scheme not in ["http", "https"]:
        return False
    # only input characters for initiatig query and/or fragments if necessary
    port_character = ":" if url_parsed.port is not None else ""
    query_character = "?" if url_parsed.query is not None else ""
    fragment_character = "#" if url_parsed.fragment is not None else ""
    url_pieces = [
        url_parsed.scheme,
        "://",
        url_parsed.host,
        port_character,
        url_parsed.port,
        url_parsed.path,
        query_character,
        url_parsed.query,
        fragment_character,
        url_parsed.fragment,
    ]
    # convert every item to a string and piece the url back together
    # to make sure it matches what was given
    url_reassembled = ""
    for url_piece in url_pieces:
        if url_piece is not None:
            url_reassembled += str(url_piece)
    # determine if parsed and reconstructed url matches original
    return str(parse_url(url)).lower() == url_reassembled.lower()


def total_amount_passed(check_status_list: list[bool]) -> tuple[int, int, float]:
    """Calculate amount of checks passed in analyze"""
    # attempt calculations for percentage of checks passed
    try:
        # calculate total amount of checks in list
        count_total = len(check_status_list)
        # count total amount of checks counted as true
        count_passed = check_status_list.count(True)
        # return tuple of checks passed, total checks, percentage of checks passed
        return (
            count_passed,
            count_total,
            (count_passed / count_total) * constants.markers.Percent_Multiplier,
        )
    # return exception of zeros when dividing by zero
    except ZeroDivisionError:
        return (0, 0, 0.0)



"""Configuration for Chasten."""

import logging
import logging.config
import logging.handlers
import sys
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import platformdirs
import requests
import yaml
from rich.logging import RichHandler
from rich.traceback import install
from urllib3.util import Url, parse_url

from chasten import constants, filesystem, output, util, validate


def configure_tracebacks() -> None:
    """Configure stack tracebacks arising from a crash to use rich."""
    install()


def user_config_dir(application_name: str, application_author: str) -> str:
    """Return the user's configuration directory using platformdirs."""
    # access the directory and then return it based on the
    # provided name of the application and the application's author
    chasten_user_config_dir_str = platformdirs.user_config_dir(
        appname=application_name,
        appauthor=application_author,
    )
    return chasten_user_config_dir_str


def configure_logging(
    debug_level: str = constants.logging.Default_Logging_Level,
    debug_dest: str = constants.logging.Default_Logging_Destination,
) -> Tuple[logging.Logger, bool]:
    """Configure standard Python logging package."""
    # use the specified logger with the specified destination
    # by dynamically constructing the function to call and then
    # invoking it with the provided debug_dest parameter
    debug_dest = debug_dest.lower()
    function_name = constants.logger.Function_Prefix + debug_dest
    configure_module = sys.modules[__name__]
    # it was possible to create the requested logger, so return it
    try:
        return (getattr(configure_module, function_name)(debug_level), True)
    # it was not possible to create the requested logger, so
    # return the default console logger as a safe alternative
    except AttributeError:
        return (configure_logging_console(debug_level), False)


def configure_logging_console(
    debug_level: str = constants.logging.Default_Logging_Level,
) -> logging.Logger:
    """Configure standard Python logging package to use rich."""
    # use the RichHandler to provide formatted
    # debugging output in the console
    logging.basicConfig(
        level=debug_level,
        format=constants.logging.Format,
        datefmt="[%X]",
        handlers=[RichHandler()],
    )
    # create a logger and then return it
    logger = logging.getLogger()
    return logger


def configure_logging_syslog(
    debug_level: str = constants.logging.Default_Logging_Level,
) -> logging.Logger:
    """Configure standard Python logging package to use syslog."""
    # use the SysLogHandler to send output to a localhost on a port
    syslog_handler = logging.handlers.SysLogHandler(
        address=(constants.server.Localhost, constants.server.Port)
    )
    logging.basicConfig(
        level=debug_level,
        format=constants.logging.Format,
        datefmt="[%X]",
        handlers=[syslog_handler],
    )
    # create a logger and then return it
    logger = logging.getLogger()
    return logger


def display_configuration_directory(
    chasten_user_config_dir_str: str, verbose: bool = False
) -> None:
    """Display information about the configuration in the console."""
    # create a visualization of the configuration directory
    chasten_user_config_dir_path = Path(chasten_user_config_dir_str)
    rich_path_tree = filesystem.create_directory_tree_visualization(
        chasten_user_config_dir_path
    )
    # display the visualization of the configuration directory
    output.opt_print_log(verbose, tree=rich_path_tree)
    output.opt_print_log(verbose, empty="")


def validate_checks_file(
    verbose: bool,
    checks_file_name: str,
    chasten_user_config_url_str: str,
    chasten_user_config_dir_str: str,
    chasten_user_config_file_str: str,
) -> Tuple[bool, bool, Dict[str, Dict[str, Any]]]:
    """Validate a checks file."""
    checks_file_validated = False
    checks_file_invalidates_entire_config = False
    # specified check file is URL
    if util.is_url(checks_file_name):
        # extract the configuration details
        (
            checks_file_extracted_valid,
            configuration_file_yaml_str,
            yaml_data_dict,
        ) = extract_configuration_details_from_config_url(parse_url(checks_file_name))
        # name of checks file is a url and thus can be used for logging
        checks_file_source = checks_file_name
    # assume check file name is a file path
    # will not support checks files being local paths
    # if config file is a URL
    elif chasten_user_config_url_str != "":
        output.logger.error(
            f"\nChecks file directive was a Path when config was a URL (given: '{checks_file_name}')\n"
        )
        checks_file_validated = False
        checks_file_invalidates_entire_config = True
        return (checks_file_validated, checks_file_invalidates_entire_config, {})
    # checks file exists in local filesystem
    elif (Path(chasten_user_config_dir_str) / Path(checks_file_name)).exists():
        # extract the configuration details
        (
            checks_file_extracted_valid,
            configuration_file_path_str,
            configuration_file_yaml_str,
            yaml_data_dict,
        ) = extract_configuration_details_from_config_dir(
            Path(chasten_user_config_dir_str), checks_file_name
        )
        # configuration path returned from extraction function can be used for logging
        checks_file_source = configuration_file_path_str
    # checks file was not valid
    else:
        output.logger.error(
            f"\nChecks file directive was not a valid Path or URL (given: '{checks_file_name}')\n"
        )
        checks_file_validated = False
        checks_file_invalidates_entire_config = True
        return (checks_file_validated, checks_file_invalidates_entire_config, {})
    # the checks file could not be extracted in a valid
    # fashion and thus there is no need to continue the
    # validation of this file or any of the other check file
    if not checks_file_extracted_valid:
        checks_file_validated = False
    # the checks file could be extract and thus the
    # function should proceed to validate a checks configuration file
    else:
        # validate checks file
        checks_file_validated = validate.validate_file(
            checks_file_source,
            configuration_file_yaml_str,
            yaml_data_dict,
            validate.JSON_SCHEMA_CHECKS,
            verbose,
        )
    return (
        checks_file_validated,
        checks_file_invalidates_entire_config,
        yaml_data_dict,
    )


def validate_configuration_files(
    config: str,
    verbose: bool = False,
) -> Tuple[
    bool, Union[Dict[str, List[Dict[str, Union[str, Dict[str, int]]]]], Dict[Any, Any]]
]:
    """Validate the configuration."""
    chasten_user_config_url_str = ""
    chasten_user_config_dir_str = ""
    chasten_user_config_file_str = ""

    if config == "":
        # there is no configuration file specified and thus
        # this function should access the platform-specific
        # configuration directory detected by platformdirs
        # detect and store the platform-specific user
        # configuration directory by default
        chasten_user_config_dir_str = user_config_dir(
            application_name=constants.chasten.Application_Name,
            application_author=constants.chasten.Application_Author,
        )
    # there is a specified configuration directory path or url;
    # this overrides the use of the configuration files that
    # may exist inside of the platform-specific directory.
    # input configuration is valid URL
    elif util.is_url(config):
        # re-parse input config so it is of type URL
        chasten_user_config_url_str = str(parse_url(config))
        output.console.print(
            ":sparkles: Configuration URL:"
            + constants.markers.Space
            + chasten_user_config_url_str
            + constants.markers.Newline
        )
        # extract the configuration details
        (
            configuration_valid,
            configuration_file_yaml_str,
            yaml_data_dict,
        ) = extract_configuration_details_from_config_url(
            parse_url(chasten_user_config_url_str)
        )
        configuration_file_source = chasten_user_config_url_str
    # input configuration exists and is valid file path
    elif Path(config).exists():
        # input configuration is a directory
        if Path(config).is_dir():
            # re-parse input config so it is of type Path
            chasten_user_config_dir_str = str(Path(config))
        # input configuration is a file
        elif Path(config).is_file():
            # re-parse input config so it is of type Path
            config_as_path = Path(config)
            # get directory containing config file
            chasten_user_config_dir_str = str(
                Path(*config_as_path.parts[: len(config_as_path.parts) - 1])
            )
            # isolate config file
            chasten_user_config_file_str = str(config_as_path.parts[-1])
        else:
            output.logger.error(
                "\nGiven configuration was a Path, but was the wrong file type.\n"
            )
            return (False, {})
        output.console.print(
            ":sparkles: Configuration directory:"
            + constants.markers.Space
            + chasten_user_config_dir_str
            + constants.markers.Newline
        )
        # optional argument if chasten_user_config_file_str is not empty
        # argument will be supplied as unpacked dict
        chasten_user_config_file_str_argument = {}
        if chasten_user_config_file_str != "":
            chasten_user_config_file_str_argument[
                "configuration_file"
            ] = chasten_user_config_file_str
        # extract the configuration details
        (
            configuration_valid,
            configuration_file_path_str,
            configuration_file_yaml_str,
            yaml_data_dict,
        ) = extract_configuration_details_from_config_dir(
            Path(chasten_user_config_dir_str), **chasten_user_config_file_str_argument
        )
        # it was not possible to extract the configuration details and
        # thus this function should return immediately with False
        # to indicate the failure and an empty configuration dictionary
        if not configuration_valid:
            return (False, {})
        # create a visualization of the user's configuration directory;
        # display details about the configuration directory in console
        display_configuration_directory(chasten_user_config_dir_str, verbose)
        configuration_file_source = chasten_user_config_dir_str
        # the configuration file does not exist and thus,
        # since config was explicit, it is not possible
        # to validate the configuration file
    else:
        output.logger.error("\nGiven configuration was not a valid Path or URL.\n")
        return (False, {})

    # Summary of the remaining steps:
    # --> Step 1: Validate the main configuration file
    # --> Step 2: Validate the one or more checks files
    # --> Step 3: If all files are valid, return overall validity
    # --> Step 3: Otherwise, return an invalid configuration
    # validate the user's configuration and display the results
    config_file_validated = validate.validate_file(
        configuration_file_source,
        configuration_file_yaml_str,
        yaml_data_dict,
        validate.JSON_SCHEMA_CONFIG,
        verbose,
    )

    # if one or more exist, retrieve the name of the checks files
    (_, checks_file_name_list) = validate.extract_checks_file_name(yaml_data_dict)
    # iteratively extract the contents of each checks file
    # and then validate the contents of that checks file
    checks_files_validated_list = []
    check_files_validated = False
    # create an empty dictionary that will store the list of checks
    overall_checks_dict: Union[
        Dict[str, List[Dict[str, Union[str, Dict[str, int]]]]], Dict[Any, Any]
    ] = {}
    # create an empty list that will store the dicts of checks
    overall_checks_list: List[Dict[str, Union[str, Dict[str, int]]]] = []
    # initialize the dictionary to contain the empty list
    overall_checks_dict[constants.checks.Checks_Label] = overall_checks_list
    for checks_file_name in checks_file_name_list:
        (
            checks_file_validated,
            checks_file_invalidates_entire_config,
            checks_file_yaml_data_dict,
        ) = validate_checks_file(
            verbose,
            checks_file_name,
            chasten_user_config_url_str,
            chasten_user_config_dir_str,
            chasten_user_config_file_str,
        )
        # checks file invalidates entire configuration
        # indicate invalid configuration
        if checks_file_invalidates_entire_config:
            return (False, {})
        # keep track of the validation of all of validation
        # records for each of the check files
        checks_files_validated_list.append(checks_file_validated)
        # add the listing of checks from the current yaml_data_dict to
        # the overall listing of checks in the main dictionary
        overall_checks_dict[constants.checks.Checks_Label].extend(checks_file_yaml_data_dict[constants.checks.Checks_Label])  # type: ignore
    # the check files are only validated if all of them are valid
    check_files_validated = all(checks_files_validated_list)
    # the files validated correctly; return an indicator to
    # show that validation worked and then return the overall
    # dictionary that contains the listing of valid checks
    if config_file_validated and check_files_validated:
        return (True, overall_checks_dict)
    # there was at least one validation error
    return (False, {})


def extract_configuration_details_from_config_dir(
    chasten_user_config_dir_str: Path,
    configuration_file: str = constants.filesystem.Main_Configuration_File,
) -> Tuple[bool, str, str, Dict[str, Dict[str, Any]]]:
    """Extract details from the configuration given a config directory.

    chasten_user_config_dir_str -- directory to search for config file
    configuration_file -- optional configuration file to specify. If not supplied, a default location will be searched
    """
    # create the name of the main configuration file
    # load the text of the main configuration file
    configuration_file_path = chasten_user_config_dir_str / configuration_file
    # the configuration file does not exist and thus
    # the extraction process cannot continue, the use of
    # these return values indicates that the extraction
    # failed and any future steps cannot continue
    if not configuration_file_path.exists():
        output.logger.error(
            f"\nFinding config or check file Path failed for {configuration_file_path}.\n"
        )
        return (False, None, None, None)  # type: ignore
    configuration_file_yaml_str = configuration_file_path.read_text()
    # load the contents of the main configuration file
    with open(str(configuration_file_path)) as user_configuration_file_text:
        (yaml_success, yaml_data) = convert_configuration_text_to_yaml(
            user_configuration_file_text.read()
        )
        # return success status, filename, file contents, and yaml parsed data upon success
        if yaml_success:
            return (
                True,
                str(configuration_file_path),
                configuration_file_yaml_str,
                yaml_data,
            )
        # return none types upon failure in yaml parsing
        else:
            output.logger.error(
                f"\nParsing YAML from config or check file Path failed for {configuration_file_path}.\n"
            )
            return (False, None, None, None)  # type: ignore


def extract_configuration_details_from_config_url(
    chasten_user_config_url: Url,
) -> Tuple[bool, str, Dict[str, Dict[str, Any]]]:
    """Extract details from the configuration given a config URL.

    chasten_user_config_url -- URL to config or checks yaml file.
    """
    # create request with given URL as source
    response = requests.get(str(chasten_user_config_url))
    # the URL response is OK
    if response.ok:
        # assume URL endpoint returns raw text
        configuration_file_yaml_str = response.text
    # the URL indicates a problem with the response
    else:
        output.logger.error(
            f"\nLoading config or check file URL failed for {chasten_user_config_url}.\n"
        )
        return (False, None, None)  # type: ignore
    (yaml_success, yaml_data) = convert_configuration_text_to_yaml(
        configuration_file_yaml_str
    )
    # return success status, filename, file contents, and yaml parsed data upon success
    if yaml_success:
        return (True, configuration_file_yaml_str, yaml_data)
    else:
        output.logger.error(
            f"\nParsing YAML from config or check file URL failed for {chasten_user_config_url}.\n"
        )
        return (False, None, None)  # type: ignore


def convert_configuration_text_to_yaml(
    configuration_file_contents_str: str,
) -> Tuple[bool, Dict[str, Dict[str, Any]]]:
    """Return details about the configuration."""
    yaml_data = None
    try:
        yaml_data = yaml.safe_load(configuration_file_contents_str)
    except Exception:
        # yaml parsing has failed and we will indicate the input is invalid
        return (False, None)  # type: ignore
    # return the file name, the textual contents of the configuration file, and
    # a dict-based representation of the configuration file
    return (True, yaml_data)



"""Store results from performing an analysis."""

import uuid
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

from pyastgrep import search as pyastgrepsearch  # type: ignore
from pydantic import BaseModel
from pydantic.types import conint

from chasten import debug

# Note: the nesting of the class definitions is from the
# bottom of this file to the top because the top-level
# object can only refer to others that already exist

# Nesting structure:

# Chasten
# --> Configuration
#     --> chastenversion
#     --> projectname
#     --> configdirectory
#     --> searchpath
#     --> debuglevel
#     --> debugdestination
#     --> checkinclude --> CheckCriterion
#     --> checkexclude --> CheckCriterion
#           --> attribute
#           --> value
#           --> confidence
# --> Source
#     --> filename
#     --> check
#         --> Check
#             --> id
#             --> name
#             --> min
#             --> max
#             --> pattern
#             --> passed
#             --> _matches
#                 --> pyastgrepsearch.Match [*]
#             --> matches
#                 --> Match
#                     --> lineno
#                     --> coloffset
#
# [*] Designates a "private" attribute that is not a part
# of the Pydantic BaseModel and is not saved to the JSON.
# This is used to record results so that the tool can
# display them if the --verbose flag is active.


class Match(BaseModel):
    """Define a Pydantic model for a Match."""

    lineno: int
    coloffset: int
    linematch: str = ""
    linematch_context: str = ""


class Check(BaseModel):
    """Define a Pydantic model for a Check."""

    id: str
    name: str
    description: str = ""
    min: Optional[conint(ge=0)] = 0  # type: ignore
    max: Optional[conint(ge=0)] = 0  # type: ignore
    pattern: str
    passed: bool
    matches: list[Match] = []
    _matches: list[pyastgrepsearch.Match] = []


class Source(BaseModel):
    """Define a Pydantic model for a Source."""

    filename: str
    _filelines: List[str] = []
    check: Union[None, Check] = None


class CheckCriterion(BaseModel):
    """Define a Pydantic model for a CheckIncludeOrExclude."""

    attribute: Union[None, str] = ""
    value: Union[None, str] = ""
    confidence: int = 0


class Configuration(BaseModel):
    """Define a Pydantic model for a Configuration."""

    chastenversion: str
    debuglevel: debug.DebugLevel
    debugdestination: debug.DebugDestination
    projectname: str
    configdirectory: Path
    searchpath: Path
    fileuuid: str = str(uuid.uuid4().hex)
    _datetime: str = str(datetime.now().strftime("%Y%m%d%H%M%S"))
    datetime: str = str(datetime.now())
    checkinclude: Union[None, CheckCriterion] = None
    checkexclude: Union[None, CheckCriterion] = None


class Chasten(BaseModel):
    """Define a Pydantic model for a Chasten result."""

    configuration: Configuration
    sources: list[Source] = []
